/***************************************************************************
 * Copyright 2012 by
 *  + Christian-Albrechts-University of Kiel
 *    + Department of Computer Science
 *      + Software Engineering Group 
 *  and others.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/

package kieker.webgui.beans.application;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.List;

import javax.annotation.PostConstruct;
import javax.faces.application.FacesMessage;
import javax.faces.application.FacesMessage.Severity;
import javax.faces.bean.ApplicationScoped;
import javax.faces.bean.ManagedBean;
import javax.faces.context.FacesContext;

import kieker.analysis.model.analysisMetaModel.MIProject;
import kieker.common.logging.Log;
import kieker.common.logging.LogFactory;
import kieker.webgui.beans.view.CurrentProjectOverviewBean;
import kieker.webgui.common.ACManager;
import kieker.webgui.common.FSManager;
import kieker.webgui.common.exception.ProjectAlreadyExistingException;

/**
 * The {@link ProjectsBean} is a JSF managed bean to manage a list with all application wide available projects. It provides methods to receive this list as well as
 * methods to add, create, rename, open and copy projects. Furthermore the state of existing projects (like the timestamp or the state of the analysis) can be
 * received via this bean. In order to realize a good abstraction, this bean should be used for any access to the projects. The necessary synchronization is achieved
 * in the {@link FSManager}.<br>
 * As this bean contains the whole list of the available projects, it is application scoped. There is no reason for multiple instances of this class.
 * 
 * @author Nils Christian Ehmke
 * @version 1.0
 */
@ManagedBean
@ApplicationScoped
public final class ProjectsBean {

	/**
	 * This is the log instance which is used by all instances of the {@link ProjectsBean}. All error and exceptions will be logged with this instance.
	 */
	private static final Log LOG = LogFactory.getLog(ProjectsBean.class);
	/**
	 * This list contains all available projects by name. It is synchronized as a simple synchronized list, as the fine grained synchronization isn't necessary here.
	 */
	private final List<String> projects = Collections.synchronizedList(new ArrayList<String>());

	/**
	 * Default constructor. <b>Do not use this constructor. This bean is JSF managed.</b>
	 */
	public ProjectsBean() {
		// No code necessary
	}

	/**
	 * This method does some initialization work after construction. <b>Do not call this method. The method is managed by the class loader.</b>
	 */
	@PostConstruct
	protected void init() {
		// Load a list with all available projects on the FS
		this.projects.addAll(FSManager.getInstance().getAllProjects());
	}

	/**
	 * This method adds a new project with the given name to the application if this is possible. In either case (this means if the project has been created, if the
	 * project could not be created for any reason) the user will be informed via the growl component. In other words: A message will be delivered within the current
	 * context. The creation of a new project means in this context that a directory with its name will be created and that an empty but valid kax-file will be
	 * produced. After creating the project (provided that he creation was successful) the projects list of the "calling" {@link CurrentProjectOverviewBean} will be
	 * updated.<br>
	 * If something should go wrong, no exception will be thrown. Everything is caught if necessary.
	 * 
	 * @param project
	 *            The name for the new project which should be added to the application.
	 */
	public void addProject(final String project) {
		try {
			// Try and use the FS-Manager to create the project atomically.
			FSManager.getInstance().addProject(project);
			// If there were no exception, everything went well. We can add the project to our list.
			this.projects.add(project);
			// Inform the user
			ProjectsBean.showMessage(FacesMessage.SEVERITY_INFO, "Project created.");
			// Update the list of the "calling" bean
			CurrentProjectOverviewBean.getCurrentInstance().updateLists();
		} catch (final IOException ex) {
			ProjectsBean.LOG.error("An error occured while creating the project.", ex);
			ProjectsBean.showMessage(FacesMessage.SEVERITY_ERROR, "An error occured while creating the project.");
		} catch (final ProjectAlreadyExistingException ex) {
			ProjectsBean.LOG.info("A project with the same name exists already.", ex);
			ProjectsBean.showMessage(FacesMessage.SEVERITY_WARN, "A project with the same name exists already.");
		}
	}

	/**
	 * This method tries to copy the project with the given name and saves it under the given new name. In either case (this means if the project has been created,
	 * if the project could not be created for any reason) the user will be informed via the growl component. In other words: A message will be delivered within the
	 * current context. After creating the project (provided that he creation was successful) the projects list of the "calling" {@link CurrentProjectOverviewBean}
	 * will be updated.<br>
	 * If something should go wrong, no exception will be thrown. Everything is caught if necessary.
	 * 
	 * @param sourceProject
	 *            The name of the source project.
	 * @param destinationProject
	 *            The name of the new (copied) project.
	 */
	public void copyProject(final String sourceProject, final String destinationProject) {
		try {
			// Try and use the FS-Manager to copy the project atomically.
			FSManager.getInstance().copyProject(sourceProject, destinationProject);
			// If there were no exception, everything went well. We can add the project to our list.
			this.projects.add(destinationProject);
			// Inform the user
			ProjectsBean.showMessage(FacesMessage.SEVERITY_INFO, "Project created.");
			// Update the list of the "calling" bean
			CurrentProjectOverviewBean.getCurrentInstance().updateLists();
		} catch (final IOException ex) {
			ProjectsBean.LOG.error("An error occured while creating the project.", ex);
			ProjectsBean.showMessage(FacesMessage.SEVERITY_ERROR, "An error occured while creating the project.");
			ex.printStackTrace();
		} catch (final ProjectAlreadyExistingException ex) {
			ProjectsBean.LOG.info("A project with the same name exists already.", ex);
			ProjectsBean.showMessage(FacesMessage.SEVERITY_WARN, "A project with the same name exists already.");
		}
	}

	/**
	 * This method can be used to open an already existing project. This means that the current state of the project on the file system is loaded into an instance of
	 * {@link MIProject}. This instance can be modified at will and for example later saved by the {@link FSManager}.
	 * 
	 * @param project
	 *            The name of the project to be opened.
	 * @return Either the model instance if the loading was successful or null otherwise. In the latter case, the user will be informed via the growl-component.
	 */
	public MIProject openProject(final String project) {
		try {
			return FSManager.getInstance().openProject(project);
		} catch (final IOException ex) {
			ProjectsBean.LOG.error("An error occured while loading the project.", ex);
			ProjectsBean.showMessage(FacesMessage.SEVERITY_ERROR, "An error occured while loading the project.");
			return null;
		}
	}

	/**
	 * This method can be used to get the current time stamp of a given project as a human readable date. If the project doesn't exist or the time stamp would be
	 * invalid "N/A" will be returned.
	 * 
	 * @param project
	 *            The project whose time stamp should be collected.
	 * @return The human readable time stamp of the project.
	 */
	public String getCurrTimeStamp(final String project) {
		// Get the current time stamp of the project
		final long timeStamp = FSManager.getInstance().getCurrTimeStamp(project);

		if (timeStamp == 0) {
			// We can assume that something went wrong
			return "N/A";
		}
		// Convert the stamp into a human readable string.
		return new Date(timeStamp).toString();

	}

	/**
	 * This method delivers all available projects as a list of string. The list itself is just a copy and can be modified as will. Keep in mind that this means also
	 * that every call to this method will create a new copy.
	 * 
	 * @return All currently available projects as a copied and modificable list.
	 */
	public List<String> getProjects() {
		synchronized (this.projects) {
			return new ArrayList<String>(this.projects);
		}
	}

	/**
	 * This method can be used to deliver the state of the analysis controller of the given project as a human readable string.
	 * 
	 * @param project
	 *            The project whose state should be delivered.
	 * @return The current state of the corresponding AnalysisController as defined by {@link ACManager#getAnalysisControllerStateString(String)}.
	 */
	public String getAnalysisControllerState(final String project) {
		return ACManager.getInstance().getAnalysisControllerStateString(project);
	}

	/**
	 * This method shows the current user a message by using the growl-component of PrimeFaces.
	 * 
	 * @param severity
	 *            The severity of the message.
	 * @param msg
	 *            The message itself.
	 */
	private static void showMessage(final Severity severity, final String msg) {
		FacesContext.getCurrentInstance().addMessage(null, new FacesMessage(severity, "", msg));
	}
}
