/***************************************************************************
 * Copyright 2012 by
 *  + Christian-Albrechts-University of Kiel
 *    + Department of Computer Science
 *      + Software Engineering Group 
 *  and others.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/

package kieker.webgui.beans.session;

import java.io.Serializable;
import java.util.Map;

import javax.faces.bean.ManagedBean;
import javax.faces.bean.SessionScoped;
import javax.faces.context.FacesContext;
import javax.servlet.http.Cookie;
import javax.servlet.http.HttpServletResponse;

/**
 * The {@link CurrentThemeBean} contains the current user theme (look and feel) of the (session) user. The default value is the "glass-x"-theme, if no other value
 * can be find within the parameters of the faces context or in the cookies of the user. Every change of the theme will also result in storing the new theme within
 * the cookies of the user.<br>
 * As the theme can be chosen by every user on his own, this class is a session scoped bean.
 * 
 * @author Nils Christian Ehmke
 * 
 * @version 1.0
 */
@ManagedBean
@SessionScoped
public final class CurrentThemeBean implements Serializable {

	/**
	 * The serial version UID.
	 */
	private static final long serialVersionUID = 2919719099912650971L;
	/**
	 * The default theme used for all users.
	 */
	private static final String DEFAULT_THEME = "glass-x";
	/**
	 * The key to identify the theme in the cookie.
	 */
	private static final String KEY_COOKIE_THEME = "theme";
	/**
	 * The key to identify the theme in the properties.
	 */
	private static final String KEY_THEME = "theme";
	/**
	 * The current theme for this session.
	 */
	private String theme;

	/**
	 * Default constructor. <b>Do not use this constructor. This bean is JSF managed.</b>
	 */
	public CurrentThemeBean() {
		this.setDefaultTheme();
		this.searchThemeInCookies();
	}

	/**
	 * This method sets the current theme of the user to the default theme. This beans that the bean tries to find the name of the theme within the parameters for
	 * primefaces. If this fail, the default theme will be used.
	 */
	private void setDefaultTheme() {
		// Get the parameters within the current context.
		final Map<String, String> params = FacesContext.getCurrentInstance().getExternalContext().getRequestParameterMap(); // NOPMD (No concurrent access)
		// Try to find the default theme within the parameters.
		if (params.containsKey(CurrentThemeBean.KEY_THEME)) {
			this.theme = params.get(CurrentThemeBean.KEY_THEME);
		} else {
			// Use the default theme.
			this.theme = CurrentThemeBean.DEFAULT_THEME;
		}
	}

	/**
	 * This method tries to search the theme within the cookies of the user.
	 */
	private void searchThemeInCookies() {
		// Try to find the cookie for the theme.
		final Map<String, Object> cookies = FacesContext.getCurrentInstance().getExternalContext().getRequestCookieMap(); // NOPMD (No concurrent access)
		if (cookies.containsKey(CurrentThemeBean.KEY_COOKIE_THEME)) {
			this.theme = ((Cookie) cookies.get(CurrentThemeBean.KEY_COOKIE_THEME)).getValue();
		}
	}

	/**
	 * Delivers the current theme.
	 * 
	 * @return The currently used theme.
	 */
	public String getTheme() {
		synchronized (this) {
			return this.theme;
		}
	}

	/**
	 * Sets the value of this bean. This beans that the stored theme will be set to the new value and that the method will try to store the name of this theme within
	 * the cookies of the user.
	 * 
	 * @param theme
	 *            The new theme to be stored within this instance.
	 */
	public void setTheme(final String theme) {
		synchronized (this) {
			this.theme = theme;

			// Set the theme cookie.
			final Cookie cookie = new Cookie(CurrentThemeBean.KEY_COOKIE_THEME, theme);
			// Try to save it for a year (maximum age)
			cookie.setMaxAge(60 * 60 * 24 * 365);

			// Deliver the cookie
			final HttpServletResponse response = (HttpServletResponse) FacesContext.getCurrentInstance().getExternalContext().getResponse();
			response.addCookie(cookie);
		}
	}
}
