/***************************************************************************
 * Copyright 2012 by
 *  + Christian-Albrechts-University of Kiel
 *    + Department of Computer Science
 *      + Software Engineering Group 
 *  and others.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/

package kieker.webgui.beans.view;

import java.io.IOException;
import java.lang.annotation.Annotation;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import javax.faces.application.FacesMessage;
import javax.faces.application.FacesMessage.Severity;
import javax.faces.bean.ManagedBean;
import javax.faces.bean.ManagedProperty;
import javax.faces.bean.ViewScoped;
import javax.faces.context.FacesContext;

import kieker.analysis.model.analysisMetaModel.MIAnalysisMetaModelFactory;
import kieker.analysis.model.analysisMetaModel.MIDependency;
import kieker.analysis.model.analysisMetaModel.MIDisplay;
import kieker.analysis.model.analysisMetaModel.MIFilter;
import kieker.analysis.model.analysisMetaModel.MIInputPort;
import kieker.analysis.model.analysisMetaModel.MIOutputPort;
import kieker.analysis.model.analysisMetaModel.MIPlugin;
import kieker.analysis.model.analysisMetaModel.MIProject;
import kieker.analysis.model.analysisMetaModel.MIProperty;
import kieker.analysis.model.analysisMetaModel.MIReader;
import kieker.analysis.model.analysisMetaModel.MIRepository;
import kieker.analysis.model.analysisMetaModel.MIRepositoryConnector;
import kieker.analysis.model.analysisMetaModel.impl.MAnalysisMetaModelFactory;
import kieker.analysis.plugin.AbstractPlugin;
import kieker.analysis.plugin.annotation.Property;
import kieker.analysis.plugin.reader.AbstractReaderPlugin;
import kieker.analysis.repository.AbstractRepository;
import kieker.common.logging.Log;
import kieker.common.logging.LogFactory;
import kieker.webgui.beans.application.ProjectsBean;
import kieker.webgui.common.FSManager;
import kieker.webgui.common.Pair;
import kieker.webgui.common.PluginFinder;
import kieker.webgui.common.exception.LibraryAlreadyExistingException;
import kieker.webgui.common.exception.NewerProjectException;
import kieker.webgui.common.exception.ProjectLoadException;

import org.primefaces.context.RequestContext;
import org.primefaces.event.FileUploadEvent;
import org.primefaces.model.UploadedFile;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;

/**
 * The {@link CurrentAnalysisEditorBean} contains the necessary data behind an instance of the analysis editor. It provides various methods to manipulate the current
 * project, as the analysis editor is the most important part of the whole application.<br>
 * The class is a JSF managed bean with view scope to make sure that one user (even in one session) can open multiple projects at a time without causing any
 * problems.
 * 
 * @author Nils Christian Ehmke
 * @version 1.0
 */
@ManagedBean
@ViewScoped
public final class CurrentAnalysisEditorBean {
	/**
	 * This is the log for errors, exceptions etc.
	 */
	private static final Log LOG = LogFactory.getLog(CurrentAnalysisEditorBean.class);
	/**
	 * This is the factory which will be used to create new components for the project.
	 */
	private final MIAnalysisMetaModelFactory factory = new MAnalysisMetaModelFactory();
	/**
	 * This is the actual model instance. It is the in-memory-model of the current (session) user.
	 */
	private MIProject project;
	/**
	 * This is the corresponding class loader to the project. It contains always the libraries within the lib-folder of the project.
	 */
	private ClassLoader classLoader;
	/**
	 * This is the corresponding class loader to the project. It contains always the correct class loader.
	 */
	private PluginFinder pluginFinder;
	/**
	 * This is the name of the stored project. It can be used as an identifier within the FS-Manager
	 */
	private String projectName;
	/**
	 * This is the time stamp of the moment, the project was loaded or last saved. It can be used to check whether the project has been modified in the meanwhile.
	 */
	private long timeStamp;
	/**
	 * This list contains the available repositories for the current project.
	 */
	private final List<Class<?>> availableRepositories = Collections.synchronizedList(new ArrayList<Class<?>>());
	/**
	 * This list contains the available filters for the current project.
	 */
	private final List<Class<?>> availableFilters = Collections.synchronizedList(new ArrayList<Class<?>>());
	/**
	 * This list contains the available readers for the current project.
	 */
	private final List<Class<?>> availableReaders = Collections.synchronizedList(new ArrayList<Class<?>>());
	/**
	 * This field contains the currently selected node (this can either be a plugin or a repository).
	 */
	private EObject selectedNode;

	@ManagedProperty(value = "#{projectsBean}")
	private ProjectsBean projectsBean;

	@ManagedProperty(value = "#{currentAnalysisEditorGraphBean}")
	private CurrentAnalysisEditorGraphBean currentAnalysisEditorGraphBean;

	private ClassAndMethodContainer classAndMethodContainer;

	/**
	 * Creates a new instance of this class.
	 */
	public CurrentAnalysisEditorBean() {
		// No code necessary
	}

	/**
	 * Getter for the attribute {@link CurrentAnalysisEditorBean#currentAnalysisEditorGraphBean}.
	 * 
	 * @return Delivers the value of the attribute.
	 */
	public CurrentAnalysisEditorGraphBean getCurrentAnalysisEditorGraphBean() {
		return this.currentAnalysisEditorGraphBean;
	}

	/**
	 * Setter for the attribute {@link CurrentAnalysisEditorBean#currentAnalysisEditorGraphBean}.
	 * 
	 * @param currentAnalysisEditorGraphBean
	 *            The new value of the attribute.
	 */
	public void setCurrentAnalysisEditorGraphBean(final CurrentAnalysisEditorGraphBean currentAnalysisEditorGraphBean) {
		this.currentAnalysisEditorGraphBean = currentAnalysisEditorGraphBean;
	}

	/**
	 * Getter for the attribute {@link CurrentAnalysisEditorBean#projectsBean}.
	 * 
	 * @return Delivers the value of the attribute.
	 */
	public ProjectsBean getProjectsBean() {
		return this.projectsBean;
	}

	/**
	 * Setter for the attribute {@link CurrentAnalysisEditorBean#projectsBean}.
	 * 
	 * @param projectsBean
	 *            The new value of the attribute.
	 */
	public void setProjectsBean(final ProjectsBean projectsBean) {
		this.projectsBean = projectsBean;
	}

	/**
	 * This method delivers the project stored in this bean.
	 * 
	 * @return The project for this user.
	 */
	public MIProject getProject() {
		synchronized (this) {
			return this.project;
		}
	}

	/**
	 * This method sets the project stored within this bean and returns the new page for the navigation - depending on the given values.
	 * 
	 * @param newName
	 *            The name of the project.
	 */
	public void setProjectName(final String newName) {
		synchronized (this) {
			// Remember the given parameters
			this.projectName = newName;
		}
	}

	/**
	 * This method initializes the bean by using the current project name to load the project. <b>Do not call this method manually. It will only be accessed by
	 * JSF.</b>
	 * 
	 * @throws ProjectLoadException
	 */
	public void initialize() throws ProjectLoadException {
		synchronized (this) {
			// Make sure that the initialization will only be done for the init request.
			if (!FacesContext.getCurrentInstance().isPostback()) {
				this.project = this.projectsBean.openProject(this.projectName);

				if (this.project != null) {
					// Remember the current time! This is important for the later comparison of the time stamps.
					this.resetTimeStamp();
					// Update the class loader and the specific classes used within various methods in this bean
					this.reloadClassLoader();
					this.reloadClassesAndMethods();
					// Add the libraries within the lib-folder to the current model
					this.addLibrariesToModel();
					// Load the available readers, filters and repositories
					this.loadToolPalette();
				}
			}
		}
	}

	private void reloadClassesAndMethods() throws ProjectLoadException {
		this.classAndMethodContainer = new ClassAndMethodContainer(this.classLoader);
	}

	/**
	 * This method loads the list of available readers, filters and repositories, using the current libraries within the model.
	 */
	private void loadToolPalette() {
		synchronized (this) {
			// Clean our tool palette
			this.availableFilters.clear();
			this.availableReaders.clear();
			this.availableRepositories.clear();

			// Make sure there is a project.
			if (this.project != null) {
				// Run through all libraries
				for (final MIDependency lib : this.project.getDependencies()) {
					this.addToToolPalette(lib);
				}
				this.addKiekerToToolPalette();
			}
		}
	}

	private void addKiekerToToolPalette() {
		synchronized (this) {
			this.addToToolPalette(FSManager.getInstance().getKiekerURL());
		}
	}

	/**
	 * This method adds all available readers, filters and repositories within the given library to the lists of this bean.
	 * 
	 * @param lib
	 *            The library used to load the plugins and repositories.
	 */
	private void addToToolPalette(final MIDependency lib) {
		synchronized (this) {
			try {
				this.addToToolPalette(FSManager.getInstance().getURL(lib, this.projectName));
			} catch (final MalformedURLException ex) {
				ex.printStackTrace();
			}
		}
	}

	/**
	 * This method adds all available readers, filters and repositories within the given library to the lists of this bean.
	 * 
	 * @param url
	 *            The library url used to load the plugins and repositories.
	 */
	private void addToToolPalette(final URL url) {
		final List<Class<?>> repositories = this.pluginFinder.getAllRepositoriesWithinJar(url);
		final List<Class<?>> plugins = this.pluginFinder.getAllPluginsWithinJar(url);
		// Now run through the available classes and add all non-abstract classes to our lists
		for (final Class<?> repository : repositories) {
			if (!Modifier.isAbstract(repository.getModifiers())) {
				this.availableRepositories.add(repository);
			}
		}
		for (final Class<?> plugin : plugins) {
			if (!Modifier.isAbstract(plugin.getModifiers())) {
				// The following cast results in the unchecked-cast-warnings, but we know that the cast should be correct.
				if (this.classAndMethodContainer.getAbstractFilterPluginClass().isAssignableFrom(plugin)) {
					this.availableFilters.add(plugin);
				} else {
					if (this.classAndMethodContainer.getAbstractReaderPluginClass().isAssignableFrom(plugin)) {
						this.availableReaders.add(plugin);
					}
				}
			}
		}
	}

	/**
	 * This method takes all libraries from the lib-folder and adds them to the in-memory-model.
	 */
	private void addLibrariesToModel() {
		synchronized (this) {
			final List<MIDependency> libs = FSManager.getInstance().getModelLibraries(this.projectName);
			// Add them, but remove all existing dependencies so far to avoid double entries. This also makes sure that the model - after it has been opened - points
			// just to valid dependencies (and to all of them).
			this.project.getDependencies().clear();
			this.project.getDependencies().addAll(libs);
		}
	}

	/**
	 * This method reloads the class loader. In other words: The class loader will always be able to load classes from the jar-files within the lib-folder of the
	 * project.
	 * 
	 * @throws ProjectLoadException
	 */
	private void reloadClassLoader() throws ProjectLoadException {
		synchronized (this) {
			this.classLoader = FSManager.getInstance().getClassLoader(this.projectName);
			try {
				this.pluginFinder = new PluginFinder(this.classLoader);
			} catch (final ClassNotFoundException ex) {
				CurrentAnalysisEditorBean.LOG.error("Could not load classes.", ex);
				throw new ProjectLoadException();
			} catch (final NullPointerException ex) {
				CurrentAnalysisEditorBean.LOG.error("Invalid class loader", ex);
				throw new ProjectLoadException();
			}
		}
	}

	/**
	 * This method sets the time stamp to the current system time.
	 */
	public void resetTimeStamp() {
		synchronized (this) {
			this.timeStamp = System.currentTimeMillis();
		}
	}

	/**
	 * This method delivers the project name stored in this bean.
	 * 
	 * @return The project name for this user.
	 */
	public String getProjectName() {
		synchronized (this) {
			return this.projectName;
		}
	}

	/**
	 * This method delivers the current time stamp.
	 * 
	 * @return The time stamp for this user.
	 */
	public long getTimeStamp() {
		synchronized (this) {
			return this.timeStamp;
		}
	}

	/**
	 * This method can be used to get the description of an {@link AbstractPlugin}- or an {@link AbstractRepository}-class. The description is read via the
	 * annotation.
	 * 
	 * @param clazz
	 *            The class whose description should be extracted.
	 * @return The description for the class or a substitute if none is available. This is in either case human readable.
	 */
	public String getDescription(final Class<?> clazz) {
		try {
			// Get the two potential annotations
			final Annotation annotationPlugin = clazz.getAnnotation(this.classAndMethodContainer.getPluginAnnotationClass());
			final Annotation annotationRepository = clazz.getAnnotation(this.classAndMethodContainer.getRepositoryAnnotationClass());

			// Now check which one of them is available
			if ((annotationPlugin == null) || ((String) this.classAndMethodContainer.getPluginDescriptionMethod().invoke(annotationPlugin, new Object[0])).isEmpty()) {
				if ((annotationRepository == null)
						|| ((String) this.classAndMethodContainer.getRepositoryDescriptionMethod().invoke(annotationRepository, new Object[0])).isEmpty()) {
					// None. Deliver a human readable substitute.
					return "No description available";
				} else {
					return (String) this.classAndMethodContainer.getRepositoryDescriptionMethod().invoke(annotationRepository, new Object[0]);
				}
			} else {
				return (String) this.classAndMethodContainer.getPluginDescriptionMethod().invoke(annotationPlugin, new Object[0]);
			}
		} catch (final IllegalAccessException ex) {
			CurrentAnalysisEditorBean.LOG.warn("Could not invoke method", ex);
			return "No description available";
		} catch (final IllegalArgumentException ex) {
			CurrentAnalysisEditorBean.LOG.warn("Could not invoke method", ex);
			return "No description available";
		} catch (final InvocationTargetException ex) {
			CurrentAnalysisEditorBean.LOG.warn("Could not invoke method", ex);
			return "No description available";
		}
	}

	public List<Annotation> getProperties(final Class<?> clazz) {
		final List<Annotation> result = new ArrayList<Annotation>();

		try {
			// Get the two potential annotations
			final Annotation annotationPlugin = clazz.getAnnotation(this.classAndMethodContainer.getPluginAnnotationClass());
			final Annotation annotationRepository = clazz.getAnnotation(this.classAndMethodContainer.getRepositoryAnnotationClass());

			final Annotation[] properties;

			// Now check which one of them is available
			if (annotationPlugin == null) {
				if (annotationRepository == null) {
					// None.
					properties = new Property[0];
				} else {
					properties = (Annotation[]) this.classAndMethodContainer.getRepositoryConfigurationMethod().invoke(annotationRepository, new Object[0]);
				}
			} else {
				properties = (Annotation[]) this.classAndMethodContainer.getPluginConfigurationMethod().invoke(annotationPlugin, new Object[0]);
			}

			for (final Annotation property : properties) {
				result.add(property);
			}

		} catch (final IllegalAccessException ex) {
			CurrentAnalysisEditorBean.LOG.warn("Could not invoke method", ex);
		} catch (final IllegalArgumentException ex) {
			CurrentAnalysisEditorBean.LOG.warn("Could not invoke method", ex);
		} catch (final InvocationTargetException ex) {
			CurrentAnalysisEditorBean.LOG.warn("Could not invoke method", ex);
		}
		return result;
	}

	/**
	 * Searches for input ports within the given class and returns them.
	 * 
	 * @param clazz
	 *            The class to be analyzed.
	 * @return A list containing the available input ports.
	 */
	public List<Annotation> getInputPorts(final Class<?> clazz) {
		final List<Annotation> result = new ArrayList<Annotation>();

		for (final Method method : clazz.getMethods()) {
			// Get the potential annotation
			final Annotation annotationPort = method.getAnnotation(this.classAndMethodContainer.getInputPortAnnotationClass());
			// Now check whether it is available
			if (annotationPort != null) {
				result.add(annotationPort);
			}
		}

		return result;
	}

	/**
	 * Searches for output ports within the given class and returns them.
	 * 
	 * @param clazz
	 *            The class to be analyzed.
	 * @return A list containing the available output ports.
	 */
	public List<Annotation> getOutputPorts(final Class<?> clazz) {
		final List<Annotation> result = new ArrayList<Annotation>();

		try {
			// Get the potential annotation
			final Annotation annotationPlugin = clazz.getAnnotation(this.classAndMethodContainer.getPluginAnnotationClass());

			// Now check whether it is available
			if (annotationPlugin != null) {
				for (final Annotation oPort : (Annotation[]) this.classAndMethodContainer.getPluginOutputPortsMethod().invoke(annotationPlugin, new Object[0])) {
					result.add(oPort);
				}
			}

		} catch (final IllegalAccessException ex) {
			CurrentAnalysisEditorBean.LOG.warn("Could not invoke method", ex);
		} catch (final IllegalArgumentException ex) {
			CurrentAnalysisEditorBean.LOG.warn("Could not invoke method", ex);
		} catch (final InvocationTargetException ex) {
			CurrentAnalysisEditorBean.LOG.warn("Could not invoke method", ex);
		}
		return result;
	}

	/**
	 * Searches for repository ports within the given class and returns them.
	 * 
	 * @param clazz
	 *            The class to be analyzed.
	 * @return A list containing the available repository ports.
	 */
	public List<Annotation> getRepositoryPorts(final Class<?> clazz) {
		final List<Annotation> result = new ArrayList<Annotation>();
		try {
			// Get the potential annotation
			final Annotation annotationPlugin = clazz.getAnnotation(this.classAndMethodContainer.getPluginAnnotationClass());

			// Now check whether it is available
			if (annotationPlugin != null) {
				for (final Annotation rPort : (Annotation[]) this.classAndMethodContainer.getPluginRepositoryPortsMethod().invoke(annotationPlugin, new Object[0])) {
					result.add(rPort);
				}
			}
		} catch (final IllegalAccessException ex) {
			CurrentAnalysisEditorBean.LOG.warn("Could not invoke method", ex);
		} catch (final IllegalArgumentException ex) {
			CurrentAnalysisEditorBean.LOG.warn("Could not invoke method", ex);
		} catch (final InvocationTargetException ex) {
			CurrentAnalysisEditorBean.LOG.warn("Could not invoke method", ex);
		}
		return result;
	}

	public List<Annotation> getDisplays(final Class<?> clazz) {
		final List<Annotation> result = new ArrayList<Annotation>();

		for (final Method method : clazz.getMethods()) {
			// Get the potential annotation
			final Annotation display = method.getAnnotation(this.classAndMethodContainer.getDisplayAnnotationClass());
			// Now check whether it is available
			if (display != null) {
				result.add(display);
			}
		}

		return result;
	}

	/**
	 * This method is the handler for the file upload. It tries to upload the given file and informs the user via the growl-component.
	 * 
	 * @param event
	 *            The upload event.
	 * @throws ProjectLoadException
	 */
	public void handleFileUpload(final FileUploadEvent event) throws ProjectLoadException {
		// Get the file from the event
		final UploadedFile file = event.getFile();

		try {
			// Use the file system manager to upload the new file
			final MIDependency lib;
			synchronized (this) {
				lib = FSManager.getInstance().uploadLibrary(file, this.projectName);
				CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_INFO, "Libary uploaded.");
				// As it seem to have worked, we can add the library to our model.
				this.project.getDependencies().add(lib);
			}
			// Update our class loader and the available plugins & repositories
			this.reloadClassLoader();
			this.addToToolPalette(lib);
		} catch (final LibraryAlreadyExistingException ex) {
			CurrentAnalysisEditorBean.LOG.info("A library with the same name exists already.", ex);
			CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_WARN, "A library with the same name exists already.");
		} catch (final IOException ex) {
			CurrentAnalysisEditorBean.LOG.error("An error occured while uploading the library.", ex);
			CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_ERROR, "An error occured while uploading the library.");
		}
	}

	/**
	 * This method delivers the available libraries of this project as a pair of strings. The first element is the name of the library, the second one te size in
	 * MiBytes as human readable string.
	 * 
	 * @return The available libraries.
	 */
	public List<Pair<String, String>> getLibraries() {
		synchronized (this) {
			return FSManager.getInstance().getLibraries(this.projectName);
		}
	}

	/**
	 * This method delivers the available reader-plugins for the current main project. The delivered plugins are never abstract.
	 * 
	 * @return A list with all readers.
	 */
	public final List<Class<?>> getAvailableReaders() {
		return this.availableReaders;
	}

	/**
	 * This method delivers the available filter-plugins for the current main project. The delivered plugins are never abstract.
	 * 
	 * @return A list with all filter.
	 */
	public final List<Class<?>> getAvailableFilters() {
		return this.availableFilters;
	}

	/**
	 * This method delivers the available repositories for the current main project. The delivered repositories are never abstract.
	 * 
	 * @return A list with all repositories.
	 */
	public final List<Class<?>> getAvailableRepositories() {
		return this.availableRepositories;
	}

	/**
	 * This method tries to save the current project and informs the user about success or fail.
	 * 
	 * @param overwriteNewerProject
	 *            This flag determines whether a newer project should be overwritten.
	 */
	public void saveProject(final boolean overwriteNewerProject) {
		synchronized (this) {
			try {
				FSManager.getInstance().saveProject(this.projectName, this.project, this.timeStamp, overwriteNewerProject);
				CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_INFO, "Project saved.");
				// Update the time stamp!
				this.resetTimeStamp();
			} catch (final IOException ex) {
				CurrentAnalysisEditorBean.LOG.error("An error occured while saving the project.", ex);
				CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_ERROR, "An error occured while saving the project.");
			} catch (final NewerProjectException ex) {
				CurrentAnalysisEditorBean.LOG.info("The project has been modified externally in the meanwhile.", ex);
				CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_WARN, "The project has been modified externally in the meanwhile.");
				// Give the user the possibility to force-save the project
				RequestContext.getCurrentInstance().execute("forceSaveDlg.show()");
			}
		}
	}

	/**
	 * This method fills the displays of the given plugin. In other words: It tries to instantiate the given class and to extract the displays. If the instantiation
	 * fails (for various reasons), the method informs the user and executes normally - without an exception.
	 * 
	 * @param clazz
	 *            The class to be used as a base.
	 * @param plugin
	 *            The plugin to be filled.
	 * @return true iff the plugin has been intialized properly.
	 */
	private boolean fillDisplays(final Class<AbstractPlugin> clazz, final MIPlugin plugin) {
		synchronized (this) {
			try {
				// Get the displays and convert them into model instances
				final List<Annotation> displays = this.getDisplays(clazz);
				for (final Annotation display : displays) {
					final MIDisplay mDisplay = this.factory.createDisplay();
					mDisplay.setName((String) this.classAndMethodContainer.getDisplayNameMethod().invoke(display, new Object[0]));
					plugin.getDisplays().add(mDisplay);
				}

				return true;
			} catch (final IllegalAccessException ex) {
				CurrentAnalysisEditorBean.LOG.error("An error occured while loading the displays of the plugin.", ex);
				CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_ERROR, "An error occured while loading the displays of the plugin.");
				return false;
			} catch (final ClassCastException ex) {
				CurrentAnalysisEditorBean.LOG.error("An error occured while loading the displays of the plugin.", ex);
				CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_ERROR, "An error occured while loading the displays of the plugin.");
				return false;
			} catch (final IllegalArgumentException ex) {
				CurrentAnalysisEditorBean.LOG.error("An error occured while loading the displays of the plugin.", ex);
				CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_ERROR, "An error occured while loading the displays of the plugin.");
				return false;
			} catch (final InvocationTargetException ex) {
				CurrentAnalysisEditorBean.LOG.error("An error occured while loading the displays of the plugin.", ex);
				CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_ERROR, "An error occured while loading the displays of the plugin.");
				return false;
			}
		}
	}

	/**
	 * This method fills the ports of the given plugin. In other words: It tries to instantiate the given class and to extract the ports. If the instantiation fails
	 * (for various reasons), the method informs the user and executes normally - without an exception.
	 * 
	 * @param clazz
	 *            The class to be used as a base.
	 * @param plugin
	 *            The plugin to be filled.
	 * @return true iff the plugin has been intialized properly.
	 */
	private boolean fillPorts(final Class<AbstractPlugin> clazz, final MIPlugin plugin) {
		synchronized (this) {
			try {
				// Get the ports
				final List<Annotation> inputPorts = this.getInputPorts(clazz);
				final List<Annotation> outputPorts = this.getOutputPorts(clazz);
				final List<Annotation> repositoryPorts = this.getRepositoryPorts(clazz);

				// Add input ports
				if (plugin instanceof MIFilter) {
					for (final Annotation inputPort : inputPorts) {
						final MIInputPort mInputPort = this.factory.createInputPort();
						mInputPort.setName((String) this.classAndMethodContainer.getInputPortNameMethod().invoke(inputPort, new Object[0]));
						mInputPort.setParent((MIFilter) plugin);
					}
				}

				// Add output ports.
				for (final Annotation outputPort : outputPorts) {
					final MIOutputPort mOutputPort = this.factory.createOutputPort();
					mOutputPort.setName((String) this.classAndMethodContainer.getOutputPortNameMethod().invoke(outputPort, new Object[0]));
					mOutputPort.setParent(plugin);
				}

				// Add repository ports.
				for (final Annotation repositoryPort : repositoryPorts) {
					final MIRepositoryConnector mConnector = this.factory.createRepositoryConnector();
					mConnector.setName((String) this.classAndMethodContainer.getRepositoryPortNameMethod().invoke(repositoryPort, new Object[0]));
					plugin.getRepositories().add(mConnector);
				}

				return true;

			} catch (final IllegalAccessException ex) {
				CurrentAnalysisEditorBean.LOG.error("An error occured while loading the ports of the plugin.", ex);
				CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_ERROR, "An errcor occured while loading the ports of the plugin.");
				return false;
			} catch (final ClassCastException ex) {
				CurrentAnalysisEditorBean.LOG.error("An error occured while loading the ports of the plugin.", ex);
				CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_ERROR, "An errcor occured while loading the ports of the plugin.");
				return false;
			} catch (final IllegalArgumentException ex) {
				CurrentAnalysisEditorBean.LOG.error("An error occured while loading the ports of the plugin.", ex);
				CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_ERROR, "An errcor occured while loading the ports of the plugin.");
				return false;
			} catch (final InvocationTargetException ex) {
				CurrentAnalysisEditorBean.LOG.error("An error occured while loading the ports of the plugin.", ex);
				CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_ERROR, "An errcor occured while loading the ports of the plugin.");
				return false;
			}
		}
	}

	/**
	 * This method fills the properties of the given repository. In other words: It tries to instantiate the given class and to extract the configuration keys. If
	 * the instantiation fails (for various reasons), the method informs the user and executes normally - without an exception.
	 * 
	 * @param clazz
	 *            The class to be used as a base.
	 * @param repository
	 *            The repository to be filled.
	 * @return true iff the repository has been intialized properly.
	 */
	private boolean fillProperties(final Class<AbstractRepository> clazz, final MIRepository repository) {
		try {
			// Get the default configuration and use it to initialize the model repository
			final List<Annotation> properties = this.getProperties(clazz);

			for (final Annotation property : properties) {
				final MIProperty mProperty = this.factory.createProperty();

				mProperty.setName((String) this.classAndMethodContainer.getPropertyNameMethod().invoke(property, new Object[0]));
				mProperty.setValue((String) this.classAndMethodContainer.getPropertyDefaultValueMethod().invoke(property, new Object[0]));

				repository.getProperties().add(mProperty);
			}

			return true;
		} catch (final IllegalAccessException ex) {
			CurrentAnalysisEditorBean.LOG.error("An error occured while loading the properties of the repository.", ex);
			CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_ERROR, "An errcor occured while loading the properties of the repository.");
			return false;
		} catch (final ClassCastException ex) {
			CurrentAnalysisEditorBean.LOG.error("An error occured while loading the properties of the repository.", ex);
			CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_ERROR, "An errcor occured while loading the properties of the repository.");
			return false;
		} catch (final IllegalArgumentException ex) {
			CurrentAnalysisEditorBean.LOG.error("An error occured while loading the properties of the repository.", ex);
			CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_ERROR, "An errcor occured while loading the properties of the repository.");
			return false;
		} catch (final InvocationTargetException ex) {
			CurrentAnalysisEditorBean.LOG.error("An error occured while loading the properties of the repository.", ex);
			CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_ERROR, "An errcor occured while loading the properties of the repository.");
			return false;
		}
	}

	/**
	 * This method fills the properties of the given plugin. In other words: It tries to instantiate the given class and to extract the configuration keys. If
	 * the instantiation fails (for various reasons), the method informs the user and executes normally - without an exception.
	 * 
	 * @param clazz
	 *            The class to be used as a base.
	 * @param plugin
	 *            The plugin to be filled.
	 * @return true iff the plugin has been intialized properly.
	 */
	private boolean fillProperties(final Class<AbstractPlugin> clazz, final MIPlugin plugin) {
		try {
			// Get the default configuration and use it to initialize the model plugin
			final List<Annotation> properties = this.getProperties(clazz);

			for (final Annotation property : properties) {
				final MIProperty mProperty = this.factory.createProperty();

				mProperty.setName((String) this.classAndMethodContainer.getPropertyNameMethod().invoke(property, new Object[0]));
				mProperty.setValue((String) this.classAndMethodContainer.getPropertyDefaultValueMethod().invoke(property, new Object[0]));

				plugin.getProperties().add(mProperty);
			}

			return true;
		} catch (final IllegalAccessException ex) {
			CurrentAnalysisEditorBean.LOG.error("An error occured while loading the properties of the plugin.", ex);
			CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_ERROR, "An errcor occured while loading the properties of the plugin.");
			return false;
		} catch (final ClassCastException ex) {
			CurrentAnalysisEditorBean.LOG.error("An error occured while loading the properties of the plugin.", ex);
			CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_ERROR, "An errcor occured while loading the properties of the plugin.");
			return false;
		} catch (final IllegalArgumentException ex) {
			CurrentAnalysisEditorBean.LOG.error("An error occured while loading the properties of the plugin.", ex);
			CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_ERROR, "An errcor occured while loading the properties of the plugin.");
			return false;
		} catch (final InvocationTargetException ex) {
			CurrentAnalysisEditorBean.LOG.error("An error occured while loading the properties of the plugin.", ex);
			CurrentAnalysisEditorBean.showMessage(FacesMessage.SEVERITY_ERROR, "An errcor occured while loading the properties of the plugin.");
			return false;
		}
	}

	/**
	 * This method adds a new repository to the current model, using the given instance of {@code Class} for it.
	 * 
	 * @param clazz
	 *            The class of the repository to be added.
	 */
	public void addRepository(final Class<AbstractRepository> clazz) {
		// Create a new instance for the model
		final MIRepository repository = this.factory.createRepository();
		repository.setClassname(clazz.getName());
		repository.setName(clazz.getSimpleName());

		final boolean check = this.fillProperties(clazz, repository);
		// Make sure that everything has been loaded correctly. If something went wrong, we don't add the component to our model
		if (check) {
			synchronized (this) {
				// Add it to the project - and to the graph
				this.project.getRepositories().add(repository);
				this.currentAnalysisEditorGraphBean.addRepository(repository);
				this.currentAnalysisEditorGraphBean.refreshGraph();
			}
		}
	}

	/**
	 * This method adds a new plugin to the current model, using the given instance of {@code Class} for it.
	 * 
	 * @param clazz
	 *            The class of the plugin to be added. This can be both, a filter or a reader.
	 */
	public void addPlugin(final Class<AbstractPlugin> clazz) {
		// Create a new instance for the model
		final MIPlugin plugin;
		if (AbstractReaderPlugin.class.isAssignableFrom(clazz)) {
			plugin = this.factory.createReader();
		} else {
			plugin = this.factory.createFilter();
		}
		plugin.setClassname(clazz.getName());
		plugin.setName(clazz.getSimpleName());

		boolean check = true;
		check &= this.fillProperties(clazz, plugin);
		check &= this.fillPorts(clazz, plugin);
		check &= this.fillDisplays(clazz, plugin);

		// Make sure that everything has been loaded correctly. If something went wrong, we don't add the component to our model
		if (check) {
			synchronized (this) {
				// Add it to the project
				this.project.getPlugins().add(plugin);

				// Add the element to the graph
				if (plugin instanceof MIReader) {
					this.currentAnalysisEditorGraphBean.addReader((MIReader) plugin);
				} else {
					this.currentAnalysisEditorGraphBean.addFilter((MIFilter) plugin);
				}
				this.currentAnalysisEditorGraphBean.refreshGraph();
			}
		}
	}

	/**
	 * This method delivers the currently selected plugin (or the currently selected repository). If nothing has been selected, null will be returned.
	 * 
	 * @return The currently selected plugin or repository.
	 */
	public EObject getSelectedPlugin() {
		synchronized (this) {
			return this.selectedNode;
		}
	}

	/**
	 * This method delivers the properties of the currently selected plugin, but it adds also the name- and the class-properties as a string to the list. The first
	 * element is always the classname, the second one is always the pluginname.
	 * 
	 * @return A list with all properties of the plugin plus the name- and class-properties. If no plugin is selected, the list of the selected repository is
	 *         delivered.
	 */
	public List<Object> getAdvancedPluginProperties() {
		synchronized (this) {
			final List<Object> result = new ArrayList<Object>();

			// Add the properties as strings
			result.add("Name");
			result.add("ClassName");

			// Get the original properties of the plugin
			if (this.selectedNode instanceof MIPlugin) {
				result.addAll(((MIPlugin) this.selectedNode).getProperties());
			} else {
				result.addAll(((MIRepository) this.selectedNode).getProperties());
			}

			return result;
		}
	}

	/**
	 * This method delivers a list containing all available filters within the model instance. It is the same as calling project.getPlugins() but instead this method
	 * returns only instances of {@link MIFilter}.
	 * 
	 * @return A list with the available filters.
	 */
	public List<MIPlugin> getFilters() {
		synchronized (this) {
			final EList<MIPlugin> plugins = this.project.getPlugins();
			final List<MIPlugin> filter = new ArrayList<MIPlugin>();

			for (final MIPlugin plugin : plugins) {
				if (plugin instanceof MIFilter) {
					filter.add(plugin);
				}
			}

			return filter;
		}
	}

	/**
	 * Delivers all available repositories.
	 * 
	 * @return A list with all repositories within the model.
	 */
	public EList<MIRepository> getRepositories() {
		synchronized (this) {
			return this.project.getRepositories();
		}
	}

	/**
	 * Delivers the name of the currently selected plugin/repository. This is only necessary for the correct renaming of the components.
	 * 
	 * @return The name of the plugin/repository.
	 */
	public String getCurrentPluginName() {
		synchronized (this) {
			if (this.selectedNode != null) {
				if (this.selectedNode instanceof MIPlugin) {
					return ((MIPlugin) this.selectedNode).getName();
				} else {
					return ((MIRepository) this.selectedNode).getName();
				}
			} else {
				return "";
			}
		}
	}

	/**
	 * This method shows the current user a message by using the growl-component of PrimeFaces.
	 * 
	 * @param severity
	 *            The severity of the message.
	 * @param msg
	 *            The message itself.
	 */
	private static void showMessage(final Severity severity, final String msg) {
		FacesContext.getCurrentInstance().addMessage(null, new FacesMessage(severity, "", msg));
	}

	/**
	 * This method initializes the modified jit-graph by delivering the necessary javascript commands to the client. It prints all current existing plugins,
	 * repositories and their connections. All further components will be added mostly by other methods.
	 */
	public void initializeGraph() {
		// Make sure that the currentAnalysisEditorGraphBean knows "this" as well.
		this.currentAnalysisEditorGraphBean.setCurrentAnalysisEditorBean(this);

		// Initialize the graph
		this.currentAnalysisEditorGraphBean.initGraph();

		// Initialize the reader, filter and repositories
		for (final MIPlugin plugin : this.project.getPlugins()) {
			if (plugin instanceof MIReader) {
				this.currentAnalysisEditorGraphBean.addReader((MIReader) plugin);
			} else {
				this.currentAnalysisEditorGraphBean.addFilter((MIFilter) plugin);
			}
		}
		for (final MIRepository repository : this.project.getRepositories()) {
			this.currentAnalysisEditorGraphBean.addRepository(repository);
		}

		// Now initialize the connections between filters
		for (final MIPlugin plugin : this.project.getPlugins()) {
			for (final MIOutputPort oPort : plugin.getOutputPorts()) {
				for (final MIInputPort iPort : oPort.getSubscribers()) {
					this.currentAnalysisEditorGraphBean.addConnection(plugin, iPort.getParent(), oPort, iPort);
				}
			}
		}

		// Repaint the graph
		this.currentAnalysisEditorGraphBean.refreshGraph();
	}

	public void nodeSelected(final EObject node) {
		synchronized (this) {
			this.selectedNode = node;
		}
	}

	public void nodeRemoved(final EObject node) {
		synchronized (this) {
			// Remove the component from the project
			if (node instanceof MIPlugin) {
				this.project.getPlugins().remove(node);

				// Remove the corresponding connections
				final List<MIInputPort> toBeRemoved = new ArrayList<MIInputPort>();
				for (final MIPlugin plugin : this.project.getPlugins()) {
					for (final MIOutputPort oPort : plugin.getOutputPorts()) {
						toBeRemoved.clear();
						for (final MIInputPort iPort : oPort.getSubscribers()) {
							if (iPort.getParent() == node) {
								toBeRemoved.add(iPort);
							}
						}
						oPort.getSubscribers().removeAll(toBeRemoved);
					}
				}
			} else {
				this.project.getRepositories().remove(node);

				// Remove the corresponding connections
				for (final MIPlugin plugin : this.project.getPlugins()) {
					for (final MIRepositoryConnector repoConn : plugin.getRepositories()) {
						if (repoConn.getRepository() == node) {
							repoConn.setRepository(null);
						}
					}
				}
			}

			// Unselect the currently selected node if it is the one which has just been removed
			if (this.selectedNode == node) {
				this.selectedNode = null; // NOPMD
			}
		}
	}

	public void edgeCreated(final MIOutputPort sourcePort, final MIInputPort targetPort) {
		sourcePort.getSubscribers().add(targetPort);
	}

	public void edgeRemoved(final MIOutputPort sourcePort, final MIInputPort targetPort) {
		sourcePort.getSubscribers().remove(targetPort);
	}

	public void edgeCreated(final MIRepositoryConnector sourcePort, final MIRepository target) {
		sourcePort.setRepository(target);
	}

	public void edgeRemoved(final MIRepositoryConnector sourcePort, final MIRepository target) {
		sourcePort.setRepository(null);
	}
}
