/***************************************************************************
 * Copyright 2012 by
 *  + Christian-Albrechts-University of Kiel
 *    + Department of Computer Science
 *      + Software Engineering Group 
 *  and others.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/

package kieker.webgui.beans.view;

import javax.faces.bean.ManagedBean;
import javax.faces.bean.ManagedProperty;
import javax.faces.bean.ViewScoped;
import javax.faces.context.FacesContext;

import kieker.analysis.display.AbstractDisplay;
import kieker.analysis.display.HtmlText;
import kieker.analysis.display.Image;
import kieker.analysis.display.PlainText;
import kieker.analysis.model.analysisMetaModel.MIProject;
import kieker.analysis.model.analysisMetaModel.MIView;
import kieker.webgui.beans.application.ProjectsBean;
import kieker.webgui.common.ACManager;
import kieker.webgui.common.Global;

/**
 * The {@link CurrentCockpitBean} contains the necessary data behind an instance of the cockpit. It provides methods to read the state of the currently
 * selected project.<br>
 * The class is a JSF managed bean with view scope to make sure that one user (even in one session) can open multiple projects at a time without causing any
 * problems.
 * 
 * @author Nils Christian Ehmke
 * @version 1.0
 */
@ManagedBean
@ViewScoped
public class CurrentCockpitBean {

	/**
	 * This is the name of the stored project. It can be used as an identifier within the FS-Manager
	 */
	private String projectName;
	/**
	 * This is the actual model instance. It is the in-memory-model of the current (session) user.
	 */
	private MIProject project;
	/**
	 * This is the currently active view.
	 */
	private MIView activeView;

	@ManagedProperty(value = "#{projectsBean}")
	private ProjectsBean projectsBean;

	/**
	 * Creates a new instance of this class.
	 */
	public CurrentCockpitBean() {
		// No code necessary
	}

	/**
	 * This method delivers the project stored in this bean.
	 * 
	 * @return The project for this user.
	 */
	public MIProject getProject() {
		synchronized (this) {
			return this.project;
		}
	}

	/**
	 * The getter for the {@link #projectsBean}-property.
	 * 
	 * @return {@link #projectsBean}
	 */
	public ProjectsBean getProjectsBean() {
		synchronized (this) {
			return this.projectsBean;
		}
	}

	/**
	 * The setter for the {@link #projectsBean}-property.
	 * 
	 * @param projectsBean
	 *            The new value for {@link #projectsBean}.
	 */
	public void setProjectsBean(final ProjectsBean projectsBean) {
		synchronized (this) {
			this.projectsBean = projectsBean;
		}
	}

	/**
	 * This method sets the project stored within this bean and returns the new page for the navigation.
	 * 
	 * @param newName
	 *            The name of the project.
	 */
	public void setProjectName(final String newName) {
		synchronized (this) {
			// Remember the given parameters
			this.projectName = newName;
		}
	}

	/**
	 * This method initializes the bean by using the current project name to load the project. <b>Do not call this method manually. It will only be accessed by
	 * JSF.</b>
	 */
	public void initalize() {
		synchronized (this) {
			// Make sure that the initialization will only be done for the init request.
			if (!FacesContext.getCurrentInstance().isPostback() && (this.projectsBean != null)) {
				// Remember the given parameters
				this.project = this.projectsBean.openProject(this.projectName);
			}
		}
	}

	/**
	 * This method delivers the project name stored in this bean.
	 * 
	 * @return The project name for this user.
	 */
	public String getProjectName() {
		synchronized (this) {
			return this.projectName;
		}
	}

	/**
	 * Delivers the current content for the given display as plain text.
	 * 
	 * @param displayName
	 *            The name of the display.
	 * @return The current content of the display, if it exists. If the display does not exist or the active view is not set 'N/A' will be returned. If the display
	 *         does exist, but the return content is not plain text, null will be returned.
	 */
	public String updatePlainTextDisplay(final String displayName) {
		final String result;

		synchronized (this) {
			if ((this.activeView != null) && (this.projectName != null)) {
				final AbstractDisplay displayObj = ACManager.getInstance().getDisplay(this.projectName, this.activeView.getName(), displayName);
				if (displayObj == null) {
					// The display does not exist
					result = "N/A";
				} else {
					if (displayObj instanceof PlainText) {
						// The display exists and is valid
						result = ((PlainText) displayObj).getText();
					} else {
						// The display exists, but is not valid
						result = null;
					}
				}
			} else {
				// The active view is not set
				result = "N/A";
			}
		}

		return result;
	}

	/**
	 * Delivers the current content for the given display as html text.
	 * 
	 * @param displayName
	 *            The name of the display.
	 * @return The current content of the display, if it exists. If the display does not exist or the active view is not set 'N/A' will be returned. If the display
	 *         does exist, but the return content is not html text, null will be returned.
	 */
	public String updateHtmlTextDisplay(final String displayName) {
		final String result;

		synchronized (this) {
			if ((this.activeView != null) && (this.projectName != null)) {
				final AbstractDisplay displayObj = ACManager.getInstance().getDisplay(this.projectName, this.activeView.getName(), displayName);
				if (displayObj == null) {
					// The display does not exist
					result = "N/A";
				} else {
					if (displayObj instanceof HtmlText) {
						// The display exists and is valid
						result = ((HtmlText) displayObj).toString();
					} else {
						// The display exists, but is not valid
						result = null;
					}
				}
			} else {
				// The active view is not set
				result = "N/A";
			}
		}

		return result;
	}

	/**
	 * Delivers the current content for the given display as Image link.
	 * 
	 * @param displayName
	 *            The name of the display.
	 * @return The current content of the display, if it exists. If the display does not exist or the active view is not set 'N/A' will be returned. If the display
	 *         does exist, but the return content is not an image, null will be returned.
	 */
	public String updateImageDisplay(final String displayName) {
		final String result;

		synchronized (this) {
			if ((this.activeView != null) && (this.projectName != null)) {
				final AbstractDisplay displayObj = ACManager.getInstance().getDisplay(this.projectName, this.activeView.getName(), displayName);
				if (displayObj == null) {
					// The display does not exist
					result = "N/A";
				} else {
					if (displayObj instanceof Image) {
						// The display exists and is valid
						result = ((Image) displayObj).toString();
					} else {
						// The display exists, but is not valid
						result = null;
					}
				}
			} else {
				// The active view is not set
				result = "N/A";
			}
		}

		return result;
	}

	/**
	 * This method clears the bean. In other words: The stored project is set to null and the method will return the page of the project overview for navigation
	 * purposes.
	 * 
	 * @return The name of the page of the project overview.
	 */
	public String clearProject() {
		synchronized (this) {
			this.projectName = null; // NOPMD
			this.project = null; // NOPMD
			this.activeView = null; // NOPMD
		}

		return Global.PAGE_PROJECT_OVERVIEW;
	}

	/**
	 * Delivers the currently active view.
	 * 
	 * @return The active view.
	 */
	public MIView getActiveView() {
		synchronized (this) {
			return this.activeView;
		}
	}

	/**
	 * Sets the active view to a new value.
	 * 
	 * @param activeView
	 *            The new active view.
	 */
	public void setActiveView(final MIView activeView) {
		synchronized (this) {
			this.activeView = activeView;
		}
	}
}
