/***************************************************************************
 * Copyright 2012 by
 *  + Christian-Albrechts-University of Kiel
 *    + Department of Computer Science
 *      + Software Engineering Group 
 *  and others.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/
package kieker.webgui.common;

import java.io.IOException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import java.util.jar.JarEntry;
import java.util.jar.JarInputStream;

import kieker.webgui.beans.view.ClassAndMethodContainer;
import kieker.webgui.common.exception.ProjectLoadException;

/**
 * This tool class can be used to find all plugins and repositories within a given jar file - assuming that the given class loader knows these jars.
 * 
 * @author Nils Christian Ehmke
 * @version 1.0
 */
public final class PluginFinder {

	private final ClassAndMethodContainer classAndMethodContainer;
	private final ClassLoader classLoader;

	/**
	 * Creates a new instance of this class, using the given class loader.
	 * 
	 * @param classLoader
	 *            The {@link ClassLoader} which will be used to load the classes.
	 * @throws ClassNotFoundException
	 * @throws ProjectLoadException
	 */
	public PluginFinder(final ClassLoader classLoader) throws ClassNotFoundException, ProjectLoadException {
		this.classLoader = classLoader;
		this.classAndMethodContainer = new ClassAndMethodContainer(classLoader);
	}

	/**
	 * This method delivers all classes which are available in the given jar and are compatible with <code>AbstractRepository</code>).
	 * 
	 * @param url
	 *            The url for the jar.
	 * @return A list containing all available repository-classes or null, if an exception occurred.
	 */
	public List<Class<?>> getAllRepositoriesWithinJar(final URL url) {
		// Get a list containing all available classes within the given jar
		final List<Class<?>> clazzes = this.getAllClassesWithinJar(url);

		List<Class<?>> result = null;

		if (clazzes != null) {
			result = new ArrayList<Class<?>>();
			for (final Class<?> clazz : clazzes) {
				// This is the cast resulting in an unchecked cast warning.
				if (clazz.isAnnotationPresent(this.classAndMethodContainer.getRepositoryAnnotationClass())
						&& this.classAndMethodContainer.getAbstractRepositoryClass().isAssignableFrom(clazz)) {
					result.add(clazz);
				}
			}
		}

		return result;
	}

	/**
	 * This method delivers all classes which are available in the given jar and have the Plugin-Annotation from the Kieker framework (And are correctly compatible
	 * with <code>AbstractPlugin</code>).
	 * 
	 * @param url
	 *            The url for the jar.
	 * @param classLoader
	 *            The class loader used to load the classes.
	 * @return A list containing all available plugin-classes or null, if an exception occurred.
	 */
	public List<Class<?>> getAllPluginsWithinJar(final URL url) {
		final List<Class<?>> clazzes = this.getAllClassesWithinJar(url);
		List<Class<?>> result = null;

		if (clazzes != null) {
			result = new ArrayList<Class<?>>();
			for (final Class<?> clazz : clazzes) {
				// This is the cast resulting in an unchecked cast warning.
				if (clazz.isAnnotationPresent(this.classAndMethodContainer.getPluginAnnotationClass())
						&& this.classAndMethodContainer.getAbstractPluginClass().isAssignableFrom(clazz)) {
					result.add(clazz);
				}
			}
		}

		return result;
	}

	/**
	 * This method delivers all classes which are available in the given jar.
	 * 
	 * @param url
	 *            The url for the jar.
	 * @param classLoader
	 *            The class loader used to load the classes.
	 * @return A list containing all available classes or null, if an exception occurred.
	 */
	private List<Class<?>> getAllClassesWithinJar(final URL url) {
		try {
			final List<Class<?>> result = new ArrayList<Class<?>>();
			// Open the jar as a stream and run through all entries within this file
			final JarInputStream stream = new JarInputStream(url.openStream());

			JarEntry jarEntry;
			while ((jarEntry = stream.getNextJarEntry()) != null) {
				try {
					// Assemble the correct name
					String name = jarEntry.toString();
					name = name.replace('/', '.');
					name = name.replace(".class", "");
					// Try to find a class with the same name and put it into our list
					final Class<?> c = this.classLoader.loadClass(name);
					result.add(c);
				} catch (final Throwable ex) { // NOPMD (Generic throwable and empty catch block) NOCS (IllegalCatchCheck)
					// Ignore error.
				}
			}
			// Don't forget to close the jar file again.
			stream.close();
			return result;
		} catch (final IOException ex) {
			ex.printStackTrace();
		}
		return null;
	}
}
